/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport.c
 * -----------
 *
 * Inter-process communication architecture
 */

#include "../../../config.h"

#include "impulse/comms/main.h"
#include "impulse/comms/transport.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/hash.h"
#include "impulse/util/thread.h"

/* 
 * The standard builtin communication transports
 */
#include "impulse/comms/transport_sock.h"


static impulse_hash transports = NULL;
#if USE_THREADS
static impulse_rwlock_static transports_lock = IMPULSE_RWLOCK_CREATE_STATIC;
#endif


void impulse_comms_init(void)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  if (transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot initialise communications subsystem",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has already been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  transports = impulse_hash_create();

#if HAVE_COMMS_SOCK
  impulse_hash_store(transports, impulse_transport_sock_tcp->name, impulse_transport_sock_tcp);
  impulse_hash_store(transports, impulse_transport_sock_udp->name, impulse_transport_sock_udp);
  impulse_hash_store(transports, impulse_transport_sock_tcp4->name, impulse_transport_sock_tcp4);
  impulse_hash_store(transports, impulse_transport_sock_udp4->name, impulse_transport_sock_udp4);
  impulse_hash_store(transports, impulse_transport_sock_tcp6->name, impulse_transport_sock_tcp6);
  impulse_hash_store(transports, impulse_transport_sock_udp6->name, impulse_transport_sock_udp6);
  impulse_hash_store(transports, impulse_transport_sock_tcpl->name, impulse_transport_sock_tcpl);
  impulse_hash_store(transports, impulse_transport_sock_udpl->name, impulse_transport_sock_udpl);
#endif
  
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


void impulse_comms_cleanup(void)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  if (!transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot cleanup communications subsystem",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has not been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  impulse_hash_destroy(transports);

  transports = NULL;
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


void impulse_comms_register(impulse_transport transport)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  impulse_hash_store(transports, transport->name, transport);
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


void impulse_comms_deregister(impulse_transport transport)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  impulse_hash_remove(transports, transport->name);
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


impulse_transport impulse_comms_lookup(impulse_unistring name)
{
  impulse_int offset = impulse_unistring_find_chr(name, (impulse_unichar)':');
  impulse_transport transport = NULL;
#if USE_THREADS
  impulse_rwlock_read(&transports_lock);
#endif
  
  if (offset < 0)
    transport = impulse_hash_recover(transports, name);
  else
  {
    impulse_unistring protocol = impulse_unistring_create_space(offset);
    impulse_unistring_substring(protocol, name, 0, offset);
    transport = impulse_hash_recover(transports, name);
    impulse_unistring_destroy(protocol);
  }
  
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
  return transport;
}


/* End of file */
