/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * colour.c
 * --------
 *
 * ROWS colour manipulation
 */

#include "../../../config.h"

#include "impulse/rows/main.h"
#include "impulse/rows/drawable_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"
#include "impulse/util/error.h"

extern ROWS_colour ROWS_colour_cie_for_drawable(void **drawable,
						int x,
						int y,
						int z)
{
  ROWS_colour_desc col;

  col.type = ROWS_colour_XYZ;
  col.data.xyz.x = x;
  col.data.xyz.y = y;
  col.data.xyz.z = z;
  
  return ROWS_colour_for_drawable(drawable, col, NULL, 0);
}

extern ROWS_colour ROWS_colour_rgb_for_drawable(void **drawable,
						int r,
						int g,
						int b)
{
  ROWS_colour_desc col;

  col.type = ROWS_colour_RGB;
  col.data.rgb.r = r;
  col.data.rgb.g = g;
  col.data.rgb.b = b;
  
  return ROWS_colour_for_drawable(drawable, col, NULL, 0);
}

extern ROWS_colour ROWS_colour_for_drawable(void             **drawable,
					    ROWS_colour_desc   colour,
					    ROWS_gamut_rgb     gamut,
					    impulse_bool       make_spot)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_colour_desc xyz, rgb;
  ROWS_colour      col;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  if (gamut == NULL)
    {
      param.colour.drawable = drawable;
      ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_GET_GAMUT_RGB])
	(&param);
      gamut = param.colour.data.get_gamut_rgb.gamut;
    }

  xyz = ROWS_colour_convert(colour, ROWS_colour_XYZ, gamut);
  rgb = ROWS_colour_convert(colour, ROWS_colour_RGB, gamut);

  col = malloc(sizeof(struct _ROWS_colour));

  col->x = xyz.data.xyz.x;
  col->y = xyz.data.xyz.y;
  col->z = xyz.data.xyz.z;

  col->r = rgb.data.rgb.r;
  col->g = rgb.data.rgb.g;
  col->b = rgb.data.rgb.b;

  col->gamut   = gamut;
  col->is_spot = make_spot;
  
  param.colour.drawable = drawable;
  param.colour.data.create.xyz = xyz;
  param.colour.data.create.rgb = rgb;
  param.colour.data.create.is_spot = 
  param.colour.data.create.data = NULL;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_CREATE])
    (&param);
  col->data = param.colour.data.create.data;
  
  ROWS_widget_unlock(drawable);

  return col;
}

/*
extern ROWS_colour ROWS_colour_create_for_drawable(void **drawable,
						   int    r,
						   int    g,
						   int    b)
{
  ROWS_colour col;
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  col = impulse_alloc(sizeof(struct _ROWS_colour));

  col->drawable = drawable;
  col->r    = r;
  col->g    = g;
  col->b    = b;

  param.colour.drawable = drawable;
  param.colour.data.create.r = r;
  param.colour.data.create.g = g;
  param.colour.data.create.b = b;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_CREATE])
    (&param);
  col->data = param.colour.data.create.data;

  ROWS_widget_unlock(drawable);

  return col;
}
*/

extern ROWS_colour ROWS_colour_create_pixel_colour(void **drawable,
						   impulse_int  pixel)
{
  ROWS_colour col;
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  col = impulse_alloc(sizeof(struct _ROWS_colour));

  col->drawable = drawable;
  col->r    = -1;
  col->g    = -1;
  col->b    = -1;

  param.colour.drawable                = drawable;
  param.colour.data.create_pixel.pixel = pixel;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_PIXEL])
    (&param);
  col->data = param.colour.data.create.data;

  ROWS_widget_unlock(drawable);

  return col;
}

void ROWS_colour_destroy(ROWS_colour c)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(c->drawable);

  private = (_p_ROWS_drawable_data **) c->drawable;

  param.colour.drawable          = c->drawable;
  param.colour.data.discard.data = c->data;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_DISCARD])
    (&param);

  ROWS_widget_unlock(c->drawable);

  impulse_free(c);
}

/*                            ======--------======                           */

/* Colour space functions */

#define NEED_GAMUT { if (gamut==NULL) { ROWS_error_raise(IMPULSE_ERROR_CRITICAL, ROWS_ERROR_INTERNAL, 0, "Colour conversion requires RGB gamut"); } }
ROWS_colour_desc ROWS_colour_convert(ROWS_colour_desc from,
				     ROWS_colour_type to,
				     ROWS_gamut_rgb   gamut)
{
  ROWS_colour_desc result;
  
  if (from.type == to)
    {
      return from;
    }

  switch (from.type)
    {
    case ROWS_colour_HSV:
    case ROWS_colour_HLS:
    case ROWS_colour_CMYK:
      /* Fixme: These colours are easier converted via RGB */
      break;

    default:
      break;
    }

  switch (to)
    {
    case ROWS_colour_HSV:
    case ROWS_colour_HLS:
    case ROWS_colour_CMYK:
      /* Fixme: These colours are easier converted via RGB */
      break;

    default:
      break;
    }
  
  if (from.type == ROWS_colour_XYZ)
    {
      result.type = to;
      
      switch (to)
	{
	case ROWS_colour_RGB:
	  NEED_GAMUT;

	  result.data.rgb.r =
	    from.data.xyz.x*gamut->invmatrix[0][0]+
	    from.data.xyz.y*gamut->invmatrix[0][1]+
	    from.data.xyz.z*gamut->invmatrix[0][2];

	  result.data.rgb.g =
	    from.data.xyz.x*gamut->invmatrix[1][0]+
	    from.data.xyz.y*gamut->invmatrix[1][1]+
	    from.data.xyz.z*gamut->invmatrix[1][2];

	  result.data.rgb.b =
	    from.data.xyz.x*gamut->invmatrix[2][0]+
	    from.data.xyz.y*gamut->invmatrix[2][1]+
	    from.data.xyz.z*gamut->invmatrix[2][2];

	  /* RGB values are pegged (not a great approximation) */
	  if (result.data.rgb.r < 0)
	    result.data.rgb.r = -1;
	  if (result.data.rgb.g < 0)
	    result.data.rgb.g = -1;
	  if (result.data.rgb.b < 0)
	    result.data.rgb.b = -1;

	  result.data.rgb.r = (result.data.rgb.r>>14);
	  result.data.rgb.g = (result.data.rgb.g>>14);
	  result.data.rgb.b = (result.data.rgb.b>>14);
	  
	  if (result.data.rgb.r > 1023)
	    result.data.rgb.r = 1024;
	  if (result.data.rgb.g > 1023)
	    result.data.rgb.g = 1024;
	  if (result.data.rgb.b > 1023)
	    result.data.rgb.b = 1024;
	  break;

	default:
	  ROWS_error_raise(IMPULSE_ERROR_CRITICAL,
			   ROWS_ERROR_INTERNAL,
			   0,
			   "Colour model not supported");
	  break;
	}
    }
  else if (to == ROWS_colour_XYZ)
    {
      result.type = to;
      
      switch (from.type)
	{
	case ROWS_colour_RGB:
	  NEED_GAMUT;

	  result.data.xyz.x =
	    from.data.rgb.r*gamut->matrix[0][0]+
	    from.data.rgb.g*gamut->matrix[0][1]+
	    from.data.rgb.b*gamut->matrix[0][2];

	  result.data.xyz.y =
	    from.data.rgb.r*gamut->matrix[1][0]+
	    from.data.rgb.g*gamut->matrix[1][1]+
	    from.data.rgb.b*gamut->matrix[1][2];

	  result.data.xyz.z =
	    from.data.rgb.r*gamut->matrix[2][0]+
	    from.data.rgb.g*gamut->matrix[2][1]+
	    from.data.rgb.b*gamut->matrix[2][2];

	  result.data.xyz.x = result.data.xyz.x>>14;
	  result.data.xyz.y = result.data.xyz.y>>14;
	  result.data.xyz.z = result.data.xyz.z>>14;
	  break;

	default:
	  ROWS_error_raise(IMPULSE_ERROR_CRITICAL,
			   ROWS_ERROR_INTERNAL,
			   0,
			   "Colour model not supported");
	}
    }
  else
    {
      return ROWS_colour_convert(ROWS_colour_convert(from,
						     ROWS_colour_XYZ,
						     gamut),
				 to,
				 gamut);
    }

  return result;
}

ROWS_gamut_rgb ROWS_colour_make_gamut_rgb (int Rx, int Ry,
					   int Gx, int Gy,
					   int Bx, int By,
					   int Cr,
					   int Cg,
					   int Cb)
{
  double matrix[3][3];
  double invmatrix[3][3];
  float  r[2], g[2], b[2];
  float  c[3];
  
  ROWS_gamut_rgb result;

  /* Hmm, yeuch, I think */
  r[0] = Rx/1024.0; r[1] = Ry/1024.0;
  g[0] = Gx/1024.0; g[1] = Gy/1024.0;
  b[0] = Bx/1024.0; b[1] = By/1024.0;
  c[0] = Cr/1024.0; c[1] = Cg/1024.0; c[2] = Cb/1024.0;

  /* [X,Y,Z] = matrix * [ R, G, B ] */
  matrix[0][0] = r[0]*c[0];
  matrix[0][1] = g[0]*c[1];
  matrix[0][2] = b[0]*c[2];
  
  matrix[1][0] = r[1]*c[0];
  matrix[1][1] = g[1]*c[1];
  matrix[1][2] = b[1]*c[2];

  matrix[2][0] = (1-r[0]-r[1])*c[0];
  matrix[2][1] = (1-g[0]-g[1])*c[1];
  matrix[2][2] = (1-b[0]-b[1])*c[2];

  /* Work out the inverse matrix ([R,G,B] = invmatrix * [X,Y,Z]) */
  {
    double factor;

    factor =
      matrix[0][0]*matrix[1][1]*matrix[2][2] +
      matrix[0][1]*matrix[1][2]*matrix[2][0] +
      matrix[0][2]*matrix[1][0]*matrix[2][1] -
      matrix[0][2]*matrix[1][1]*matrix[2][0] -
      matrix[0][0]*matrix[1][2]*matrix[2][1] -
      matrix[0][1]*matrix[1][0]*matrix[2][2];

    invmatrix[0][0] =
      (matrix[2][2]*matrix[1][1] - matrix[1][2]*matrix[2][1]) /
      factor;
    invmatrix[0][1] =
      (matrix[0][2]*matrix[2][1] - matrix[2][2]*matrix[0][1]) /
      factor;
    invmatrix[0][2] =
      (matrix[1][2]*matrix[0][1] - matrix[0][2]*matrix[1][1]) /
      factor;

    invmatrix[1][0] =
      (matrix[1][2]*matrix[2][0] - matrix[2][2]*matrix[1][0]) /
      factor;
    invmatrix[1][1] =
      (matrix[2][2]*matrix[0][0] - matrix[0][2]*matrix[2][0]) /
      factor;
    invmatrix[1][2] =
      (matrix[0][2]*matrix[1][0] - matrix[1][2]*matrix[0][0]) /
      factor;

    invmatrix[2][0] =
      (matrix[2][1]*matrix[1][0] - matrix[1][1]*matrix[2][0]) /
      factor;
    invmatrix[2][1] =
      (matrix[0][1]*matrix[2][0] - matrix[2][1]*matrix[0][0]) /
      factor;
    invmatrix[2][2] =
      (matrix[1][1]*matrix[0][0] - matrix[0][1]*matrix[1][0]) /
      factor;
  }

  /* Fill in the important fields */
  result = malloc(sizeof(struct _ROWS_gamut_rgb));

  {
    int x,y;

    for (x=0; x<3; x++)
      for (y=0; y<3; y++)
	{
	  result->matrix[x][y]    = matrix[x][y]*16384;
	  result->invmatrix[x][y] = invmatrix[x][y]*16384;
	}
  }

  /* Fill in the informative fields */
  result->r[0] = Rx; result->r[1] = Ry;
  result->g[0] = Gx; result->g[1] = Gy;
  result->b[0] = Bx; result->b[1] = By;

  result->C[0] = Cr;
  result->C[1] = Cg;
  result->C[2] = Cb;
  
  return result;
}

/* End of file */
