/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * screen.c
 * --------
 *
 * An implementation of drawables for drawing on screen
 */

#include "../../../config.h"

#include "impulse/rows/screen_int.h"

#if HAVE_GUI_XWIN
#include "impulse/rows/screen_xwin.c"
#else
#error "Screens not implemeneted on this platform (bummer!)"
#endif

/*                            ======--------======                           */

static void ROWS_screen_setup(ROWS_widget *widget,
			      ROWS_child child);
int            screen_ID         = -1;
int            screen_type       = -1;
ROWS_gamut_rgb ROWS_screen_gamut = NULL;

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_screen_create(ROWS_child child)
{
  ROWS_widget *wid;

  if (screen_ID == -1)
    screen_ID   = ROWS_drawable_next_identifier();

  wid = ROWS_screen_initialise(screen_ID, child);

  return *wid;
}

ROWS_widget *ROWS_screen_initialise(int id,
				    ROWS_child child)
{
  ROWS_widget *widget;

  if (screen_ID   == -1)
    screen_ID   = ROWS_drawable_next_identifier();
  if (screen_type == -1)
    screen_type = ROWS_widget_next_type();

  widget = ROWS_drawable_initialise(id);
  ROWS_widget_set_type(widget, screen_ID, screen_type);

  ROWS_screen_setup(widget, child);
  
  return widget;
}

int ROWS_screen_next_identifier(void)
{
  if (screen_ID==-1)
    screen_ID = ROWS_drawable_next_identifier();

  return screen_ID + 1;
}

ROWS_screen ROWS_screen_cast(void **widget)
{
  if (screen_ID   == -1)
    screen_ID   = ROWS_drawable_next_identifier();
  if (screen_type == -1)
    screen_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, screen_ID, screen_type);
}

void ROWS_screen_colour_get_gamut_rgb(ROWS_drawable_fparam *param)
{
  if (ROWS_screen_gamut == NULL)
    {
      /*
       * Huh? We should have set this up when the screen object was
       * created...
       */

      ROWS_error_raise(IMPULSE_ERROR_CRITICAL, ROWS_ERROR_INTERNAL, 0,
		       "Programmer is a spoon: tried to get screen gamut before it was created");
    }

  param->colour.data.get_gamut_rgb.gamut = ROWS_screen_gamut;
}

void ROWS_screen_setup(ROWS_widget *widget,
		       ROWS_child child)
{
  ROWS_drawable_brush_calls  *layer;

  if (ROWS_screen_gamut == NULL)
    {
      /*
       * These values are from 'Computer Graphics: Principles &
       * Practice' (Foley, van Dam, Feiner & Hughes) and are the
       * values for a 'typical monitor', whatever that is. Colour
       * gamuts for other monitors welcome...
       */
      ROWS_screen_gamut = ROWS_colour_make_gamut_rgb(/* Col X,Y coordinates */
					     625 /* 0.61 */, 358 /* 0.35 */,
					     297 /* 0.29 */, 604 /* 0.59 */,
					     154 /* 0.15 */, 65  /* 0.06 */,
					     /* Chromacity */
					     /*
					      * Total guess. Assuming
					      * white is at (1/3, 1/3, 1/3)
					      */
					     298 /* 0.29 */,
					     363 /* 0.35 */,
					     363 /* 0.35 */);
    }
  
  layer = ROWS_drawable_get_abstraction_layer(ROWS_drawable_cast(*widget));
  
  layer->function[ROWS_DRAWABLE_CLEAR_AREA] =
    ROWS_screen_clear_area;
  layer->function[ROWS_DRAWABLE_SYNC] =
    ROWS_screen_sync;
  layer->function[ROWS_DRAWABLE_DRAW_BITMAP] =
    ROWS_screen_draw_bitmap;

  layer->function[ROWS_DRAWABLE_CLIP_PUSH] =
    ROWS_screen_clip_push;
  layer->function[ROWS_DRAWABLE_CLIP_POP] =
    ROWS_screen_clip_pop;
  layer->function[ROWS_DRAWABLE_CLIP_CLEAR] =
    ROWS_screen_clip_clear;
  layer->function[ROWS_DRAWABLE_CLIP_UNION] =
    ROWS_screen_clip_union;
  layer->function[ROWS_DRAWABLE_CLIP_INTERSECT] =
    ROWS_screen_clip_intersect;
  layer->function[ROWS_DRAWABLE_CLIP_SUBTRACT] =
    ROWS_screen_clip_subtract;
  layer->function[ROWS_DRAWABLE_CLIP_SET] =
    ROWS_screen_clip_set;
  layer->function[ROWS_DRAWABLE_CLIP_BOUNDS] =
    ROWS_screen_clip_bounds;
  layer->function[ROWS_DRAWABLE_CLIP_MOVE] =
    ROWS_screen_clip_move;
  layer->function[ROWS_DRAWABLE_CLIP_COPY] =
    ROWS_screen_clip_copy;

  layer->function[ROWS_DRAWABLE_BRUSH_USE_PRIVATE_GC] =
    ROWS_screen_brush_use_private_gc;
  layer->function[ROWS_DRAWABLE_BRUSH_SET_GC] = 
    ROWS_screen_brush_set_gc;

  layer->function[ROWS_DRAWABLE_BRUSH_SET_LINEWIDTH] = 
    ROWS_screen_brush_set_linewidth;
  layer->function[ROWS_DRAWABLE_BRUSH_GET_FGPIXEL] = 
    ROWS_screen_brush_get_fgpixel;
  layer->function[ROWS_DRAWABLE_BRUSH_SET_FGCOLOUR] = 
    ROWS_screen_brush_set_fgcolour;
  layer->function[ROWS_DRAWABLE_BRUSH_SET_BGCOLOUR] = 
    ROWS_screen_brush_set_bgcolour;

  layer->function[ROWS_DRAWABLE_BRUSH_DRAW_POINT] = 
    ROWS_screen_brush_draw_point;
  layer->function[ROWS_DRAWABLE_BRUSH_DRAW_LINE] = 
    ROWS_screen_brush_draw_line;
 layer->function[ROWS_DRAWABLE_BRUSH_DRAW_RECTANGLE] = 
    ROWS_screen_brush_draw_rectangle;
  layer->function[ROWS_DRAWABLE_BRUSH_DRAW_ARC] = 
    ROWS_screen_brush_draw_arc;
  layer->function[ROWS_DRAWABLE_BRUSH_FILL_RECTANGLE] = 
    ROWS_screen_brush_fill_rectangle;
  layer->function[ROWS_DRAWABLE_BRUSH_FILL_ARC] =
    ROWS_screen_brush_fill_arc;
  
  layer->function[ROWS_DRAWABLE_BRUSH_SET_FONT] =
    ROWS_screen_brush_set_font;
  layer->function[ROWS_DRAWABLE_BRUSH_SET_FONT_SIZE] =
    ROWS_screen_brush_set_font_size;
  layer->function[ROWS_DRAWABLE_BRUSH_FONT_PAINT] =
    ROWS_screen_brush_font_paint;
  layer->function[ROWS_DRAWABLE_BRUSH_FONT_MEASURE] =
    ROWS_screen_brush_font_measure;

  layer->function[ROWS_DRAWABLE_COLOUR_GET_GAMUT_RGB] =
    ROWS_screen_colour_get_gamut_rgb;
  layer->function[ROWS_DRAWABLE_COLOUR_CREATE] =
    ROWS_screen_colour_create;
  layer->function[ROWS_DRAWABLE_COLOUR_PIXEL] =
    ROWS_screen_colour_pixel;
  layer->function[ROWS_DRAWABLE_COLOUR_DISCARD] =
    ROWS_screen_colour_discard;

  layer->function[ROWS_DRAWABLE_BRUSH_CREATE] =
    ROWS_screen_brush_create;
  layer->function[ROWS_DRAWABLE_BRUSH_DESTROY] =
    ROWS_screen_brush_destroy;
  
  ROWS_screen_setup_int(widget, child);
}

/* End of file */
