/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * module.c
 * --------
 *
 * Impulse runtime dynamic library loading
 */

/*********************************************************************
 * 
 * VIW (Very Important Warning)
 * 
 * No functions used impulse_module_* must *EVER* call
 * 
 *  impulse_unistring_to_8bit
 *  impulse_unistring_from_8bit
 *  impulse_text_get
 * 
 * All of these functions make use of dynamically loaded modules,
 * and thus would have huge potential for infinite loops.....
 * 
 * Unfortunately this means that error messages in this module
 * can not be internationalised. In fact do we really need error 
 * messages here, seeing as the caller will have to deal with the
 * NULL return values and will likely display a suitable message
 * itself.
 * 
 ********************************************************************/

#include "../../../config.h"

#include "impulse/util/module.h"
#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/hash.h"
#include "impulse/util/thread.h"
#include "impulse/util/event.h"

/*                            ======--------======                           */

#if HAVE_MODULE_DL
# include <dlfcn.h>
#elif HAVE_MODULE_DLD
# include <dld.h>
#elif HAVE_MODULE_WIN32
# include <windows.h>
#elif HAVE_MODULE_SHL
# include <dl.h>
#else
#error "Doh, dynamic modules don't seem to be supported on your platform"
#endif

/*                            ======--------======                           */


struct _impulse_module
{
  impulse_unistring name;
  impulse_unistring filename;
  void *handle;
  impulse_int refs;
  
  impulse_int ndeps;
  struct _impulse_module **deps;
  
  impulse_int timer;
  impulse_bool timing;
};

/*                            ======--------======                           */

static impulse_bool autoclean = 0;
static impulse_hash modules = NULL;
#if USE_THREADS
static impulse_mutex_static module_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif

/*                            ======--------======                           */

#if 0
static void impulse_module_error_int(impulse_unistring desc);
#endif

static impulse_module impulse_module_create_int(impulse_unistring name);
static void impulse_module_destroy_int(impulse_module module,
				       impulse_bool deps);

static void impulse_module_cleaner_int(void *);
static impulse_bool impulse_module_timer_int(impulse_event *event,
					     void *data);

static impulse_module impulse_module_load_int(impulse_unistring filename);
static void impulse_module_unload_int(impulse_module module);
static void *impulse_module_symbol_int(impulse_module module,
				       impulse_unistring name);


/*                            ======--------======                           */

#if 0
static void impulse_module_error_int(impulse_unistring desc)
{
#if HAVE_MODULE_SHL
  impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
#else
# if HAVE_MODULE_DL
#  if HAVE_MODULE_DL_DLERROR
  impulse_unistring msg = impulse_unistring_create_utf(dlerror(), 8);
#  else
  impulse_unistring msg = impulse_unistring_create_utf("Unknown error", 8);
#  endif
# elif HAVE_MODULE_DLD
  impulse_unistring msg = impulse_unistring_create_utf(dld_strerror(dld_errno), 8);
# elif HAVE_MODULE_WIN32
  impulse_unistring msg = impulse_unistring_create_utf("Unknown error", 8);
# endif
  impulse_error_raise(IMPULSE_ERROR_WARNING, desc, msg);
  impulse_unistring_destroy(msg);
#endif
}
#endif


static void impulse_module_cleaner_int(void *data)
{
  impulse_module module = (impulse_module)data;
  
  /* 
   * Since we are unloading all modules at once, we don't have
   * to worry about doing dependancies explicitly. This is
   * just as well, cos unloading dependancies hits the hash
   * table which is not allowed from within cleaner functions.
   */
  module->ndeps = 0;
  impulse_module_unload_int(module);
}


static impulse_bool impulse_module_timer_int(impulse_event *event,
					     void *data)
{
  impulse_module module = (impulse_module)data;

  /*
   * NB. THis is the only private interface which needs to lock
   * the module mutex. With all other private interfaces, the
   * mutex is already locked by the public interface
   */
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  if (!module->refs)
  {
    impulse_hash_remove(modules, module->filename);
    impulse_module_unload_int(module);
  }
  else
    module->timing = 0;
  
#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
  return 0;
}


static impulse_module impulse_module_create_int(impulse_unistring name)
{
  impulse_module module = NULL;
  
  while (!module)
  {
    impulse_unistring filename = impulse_unistring_create(name);
    
    impulse_unistring_append_utf(filename, ".so", 8);

    if ((module = impulse_hash_recover(modules, filename)))
      module->refs++;
    else
    {
      if ((module = impulse_module_load_int(filename)))
	impulse_hash_store(modules, module->filename, module);
    }
    
    break;
  }
  
  return module;
}


static void impulse_module_destroy_int(impulse_module module,
				       impulse_bool deps)
{
  /* Don't do anything else if the module is still in use */
  if (--module->refs)
    return;
  
  /* Don't postpone unloading if this is a dependancy library */
  if (autoclean && !deps)
  {
    /* See if a timeout is already scheduled */
    if (!module->timing)
    {
      if (!module->timer)
	module->timer = impulse_event_timer();
      
      impulse_event_register_ticker(module->timer,
				    60,
				    0,
				    impulse_module_timer_int,
				    module);
      module->timing = 1;
    }
    else
      ; /* Fix me: we should be extending the timeout to the full 60 seconds here */
  }
  else
  {
    impulse_hash_remove(modules, module->filename);
    impulse_module_unload_int(module);
  }
}


static impulse_module impulse_module_load_int(impulse_unistring filename)
{
  impulse_module module = impulse_alloc(sizeof(struct _impulse_module));
  impulse_int offset1 = -1;
  impulse_int offset2 = -1;
  impulse_char *cfile = NULL;
  
  module->filename = impulse_unistring_create(filename);
  module->handle = NULL;
  module->refs = 1;
  module->ndeps = 0;
  module->deps = NULL;
  module->timer = 0;
  module->timing = 0;
  
  
  offset1 = impulse_unistring_rfind_chr(filename, (impulse_unichar)'/');
  if (offset1 == -1)
    module->name = impulse_unistring_create(filename);
  else
  {
    offset2 = impulse_unistring_find_from_chr(filename, (impulse_unichar)'.', offset1);
    if (offset2 == -1)
      offset2 = filename->length;
    
    module->name = impulse_unistring_create_space(offset2-(offset1+1));
    impulse_unistring_substring(module->name, filename, offset1+1, offset2);
  }

  cfile = impulse_unistring_to_utf(module->filename, 8, 0);

#if HAVE_MODULE_DL
  module->handle = dlopen(cfile, RTLD_NOW | RTLD_GLOBAL);
#elif HAVE_MODULE_DLD
  module->handle = dld_link(cfile);
#elif HAVE_MODULE_WIN32
  module->handle = LoadLibrary(cfile);
#elif HAVE_MODULE_SHL
  module->handle = shl_load(cfile, BIND_IMMEDIATE | BIND_NOSTART, 0L);
#endif

  impulse_free(cfile);
  
  if (!module->handle)
  {
    impulse_unistring_destroy(module->name);
    impulse_unistring_destroy(module->filename);
    impulse_free(module);
    return NULL;
  }
  
  return module;
}


static void impulse_module_unload_int(impulse_module module)
{
  impulse_int i;

#if HAVE_MODULE_DL
  dlclose(module->handle);
#elif HAVE_MODULE_DLD
  impulse_char *cstr = impulse_unistring_to_utf(module->filename, 8, 0);
  dld_unlink_by_file(cstr, 1);
  impulse_free(cstr);
#elif HAVE_MODULE_WIN32
  FreeLibrary(module->handle);
#elif HAVE_MODULE_SHL
  shl_unload(module->handle);
#endif

  impulse_unistring_destroy(module->name);
  impulse_unistring_destroy(module->filename);

  /* Unload dependancies */
  for (i = 0 ; i < module->ndeps ; i++)
    impulse_module_destroy_int(module->deps[i], 1);
  
  impulse_free(module);
}

  
static void *impulse_module_symbol_int(impulse_module module,
				       impulse_unistring name)
{
  impulse_char *cstr = impulse_unistring_to_utf(name, 8, 0);
  void *sym = NULL;

#if HAVE_MODULE_DL
  sym = dlsym(module->handle, cstr);
#elif HAVE_MODULE_DLD
  sym = dld_get_func(cstr);
#elif HAVE_MODULE_WIN32
  sym = GetProcAddress(module->handle, cstr);
#elif HAVE_MODULE_SHL
  if (shl_findsym(module->handle, cstr, TYPE_UNDEFINED, &sym) < 0)
    sym = NULL;
#endif
  impulse_free(cstr);
  
  if (!sym)
  {
    impulse_unistring pref = impulse_unistring_create_utf("_", 8);

    impulse_unistring_prepend(pref, module->name);
    impulse_unistring_append(pref, name);
    cstr = impulse_unistring_to_utf(pref, 8, 0);

#if HAVE_MODULE_DL
    sym = dlsym(module->handle, cstr);
#elif HAVE_MODULE_DLD
    sym = dld_get_func(cstr);
#elif HAVE_MODULE_WIN32
    sym = GetProcAddress(module->handle, cstr);
#elif HAVE_MODULE_SHL
    if (shl_findsym(module->handle, cstr, TYPE_UNDEFINED, &sym) < 0)
      sym = NULL;
#endif

    impulse_free(cstr);
    impulse_unistring_destroy(pref);
  }
  
  return sym;
}

  
/*                            ======--------======                           */

  /* 
   * Public Interface below this point
   */
  
#if HAVE_MODULE_DLD
extern impulse_unistring prog_path;
#endif


void impulse_module_init(void)
{
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  modules = impulse_hash_create();
  impulse_hash_cleaner(modules,
		       impulse_module_cleaner_int);

#if HAVE_MODULE_DLD
  {
    impulse_char *cstr = impulse_unistring_to_utf(prog_path, 8, 0);
    dld_init(cstr);
    impulse_free(cstr);
  }
#endif

#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
}


void impulse_module_cleanup(void)
{
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  impulse_hash_destroy(modules);
  
#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
}


impulse_module impulse_module_create(impulse_unistring name)
{
  impulse_module mod = NULL;

#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  mod = impulse_module_create_int(name);

#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
  return mod;
}


void impulse_module_destroy(impulse_module module)
{
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif

  impulse_module_destroy_int(module, 0);

#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
}


void *impulse_module_symbol(impulse_module module,
			    impulse_unistring name)
{
  void *sym = NULL;
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  sym = impulse_module_symbol_int(module, name);
  
#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
  return sym;
}


void impulse_module_autoclean(impulse_bool on)
{
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  autoclean = on;
  
#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
}


/* End of file */
