/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * stack.c
 * -------
 *
 * Double headed stacks
 */

#include "../../../config.h"

#include "impulse/util/memory.h"
#include "impulse/util/stack.h"

/*                            ======--------======                           */

struct _impulse_stack_bucket
{
  void *val;
  struct _impulse_stack_bucket *next;
};


typedef struct _impulse_stack_bucket *impulse_stack_bucket;

struct _impulse_stack
{
  impulse_stack_bucket head;
  impulse_stack_bucket iters;
  
  impulse_uint count;
  
  impulse_stack_cleaner_func cleaner;
};


struct _impulse_stack_iterator
{
  impulse_stack stack;
  impulse_stack_bucket current;
};


typedef struct _impulse_stack_iterator *impulse_stack_iterator;

impulse_stack_iterator impulse_stack_iterator_create(impulse_stack stack);
void impulse_stack_iterator_destroy(void *data);
impulse_bool impulse_stack_iterator_eof(void *data);
void impulse_stack_iterator_reset(void *data);
void *impulse_stack_iterator_next(void *data);

static struct _impulse_iterator_funcs iterator_funcs = 
{
  impulse_stack_iterator_destroy,
  impulse_stack_iterator_eof,
  impulse_stack_iterator_reset,
  impulse_stack_iterator_next
};

/*                            ======--------======                           */

impulse_stack impulse_stack_create(void)
{
  impulse_stack stack = impulse_alloc(sizeof(struct _impulse_stack));
  
  stack->head = NULL;
  stack->iters = NULL;
  stack->count = 0;
  stack->cleaner = NULL;
  return stack;
}


void impulse_stack_destroy(impulse_stack stack)
{
  impulse_stack_bucket current = stack->iters, next;
  impulse_stack_clear(stack);
  
  while (current)
  {
    impulse_stack_iterator si = (impulse_stack_iterator)current->val;
    si->stack = NULL;
    next = current->next;
    impulse_free(current);
    current = next;
  }
  
  impulse_free(stack);
}


void impulse_stack_cleaner(impulse_stack stack,
			   impulse_stack_cleaner_func func)
{
  stack->cleaner = func;
}


impulse_iterator impulse_stack_elements(impulse_stack stack)
{
  impulse_stack_iterator it = impulse_stack_iterator_create(stack);
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs, it);
  return iterator;
}


impulse_bool impulse_stack_empty(impulse_stack stack)
{
  return stack->count == 0;
}


impulse_uint impulse_stack_count(impulse_stack stack)
{
  return stack->count;
}


void impulse_stack_push(impulse_stack stack,
			void *data)
{
  impulse_stack_bucket temp = impulse_alloc(sizeof(struct _impulse_stack_bucket));

  temp->val = data;
  temp->next = stack->head;
  stack->head = temp;
  stack->count++;
}


void *impulse_stack_pop(impulse_stack stack)
{
  impulse_stack_bucket temp = stack->head;
  impulse_stack_bucket current = stack->iters;
  void *data = NULL;
  
  if (!stack->head)
    return NULL;

  /* Update any iterators pointing to the element to be popped */
  while (current)
  {
    impulse_stack_iterator si = (impulse_stack_iterator)current->val;
    if (si->current == stack->head)
      si->current = stack->head->next;
    current = current->next;
  }

  data = temp->val;
  stack->head = stack->head->next;
  impulse_free(temp);
  stack->count--;
  return data;
}


void *impulse_stack_peek(impulse_stack stack)
{
  if (!stack->head)
    return NULL;
  return stack->head->val;
}


void impulse_stack_clear(impulse_stack stack)
{
  impulse_stack_bucket temp = stack->head, next = NULL;
  impulse_stack_bucket current = stack->iters;

  while (temp)
  {
    next = temp->next;
    if (stack->cleaner)
      (*stack->cleaner)(temp->val);
    impulse_free(temp);
    temp = next;
  }
  stack->count = 0;
  
  /* Update all stack iterators */
  while (current)
  {
    impulse_stack_iterator si = (impulse_stack_iterator)current->val;
    si->current = NULL;
    current = current->next;
  }
}


impulse_bool impulse_stack_contains(impulse_stack stack,
				    void *data)
{
  impulse_stack_bucket current = stack->head;
  while (current)
  {
    if (current->val == data)
      return 1;
    current = current->next;
  }
  return 0;
}


impulse_stack_iterator impulse_stack_iterator_create(impulse_stack stack)
{
  impulse_stack_iterator it = impulse_alloc(sizeof(struct _impulse_stack_iterator));
  impulse_stack_bucket temp = impulse_alloc(sizeof(struct _impulse_stack_bucket));
  
  it->stack = stack;
  
  /* Put us in the list of stack iterators */
  temp->next = stack->iters;
  temp->val = it;
  stack->iters = temp;
  it->current = NULL;
  
  impulse_stack_iterator_reset(it);
  
  return it;
}


void impulse_stack_iterator_destroy(void *data)
{
  impulse_stack_iterator it = (impulse_stack_iterator)data;
  
  /* Remove us from the list of active stack iterators */
  if (it->stack)
  {
    impulse_stack_bucket current = it->stack->iters, prev = NULL;
    while (current)
    {
      if (current->val == it)
      {
        if (prev)
          prev->next = current->next;
        else
          it->stack->iters = current->next;
        impulse_free(current);
        break;
      }
      prev = current;
      current = current->next;
    }
  }
  
  impulse_free(it);
}


impulse_bool impulse_stack_iterator_eof(void *data)
{
  impulse_stack_iterator it = (impulse_stack_iterator)data;
  return it->current == NULL;
}


void impulse_stack_iterator_reset(void *data)
{
  impulse_stack_iterator it = (impulse_stack_iterator)data;
  if (it->stack)
    it->current = it->stack->head;
  else
    it->current = NULL;
}


void *impulse_stack_iterator_next(void *data)
{
  impulse_stack_iterator it = (impulse_stack_iterator)data;
  void *val = NULL;
  
  if (it->current)
  {
    val = it->current->val;
    it->current = it->current->next;
  }
    
  return val;
}


/* End of file */
