/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * thread.c
 * --------
 *
 * Impulse threading routines
 */

#include "../../../config.h"

#include "impulse/util/setup.h"
#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"

#include <stdlib.h>
#include <stdio.h>

#if USE_THREADS

/*                            ======--------======                           */
#define THREAD_ERROR(x) \
if ((x) < 0) \
{ \
  impulse_unistring desc = impulse_unistring_create_utf("Multi-threading safety violated", 8); \
  impulse_error_syscall(IMPULSE_ERROR_EMERGENCY, desc); \
  impulse_unistring_destroy(desc); \
}

#if HAVE_THREADS_POSIX
# include "impulse/util/thread_posix.c"
#elif HAVE_THREADS_PTH
# include "impulse/util/thread_pth.c"
#elif HAVE_THREADS_SUN
# include "impulse/util/thread_sun.c"
#elif HAVE_THREADS_WIN32
# include "impulse/util/thread_win32.c"
#else
# error "Oh dear - unsupported thread library configured. Reboot programmer and recompile"
#endif



struct _p_impulse_thread_message {
  void                             *data;
  struct _p_impulse_thread_message *next;
};

struct _impulse_message_queue {
  impulse_mutex mutex;
  impulse_cond condition;

  struct _p_impulse_thread_message *first;
  struct _p_impulse_thread_message *last;
};


impulse_message_queue impulse_message_queue_create(void)
{
  impulse_message_queue queue = impulse_alloc(sizeof(struct _impulse_message_queue));
  
  queue->mutex = impulse_mutex_create();
  queue->condition = impulse_cond_create();
  queue->first = NULL;
  queue->last  = NULL;

  return queue;
}

void impulse_message_queue_destroy(impulse_message_queue queue)
{
  struct _p_impulse_thread_message *msg = queue->first;

  while (msg)
  {
    struct _p_impulse_thread_message *next = msg->next;
    
    impulse_free(msg->data);
    impulse_free(msg);
    msg = next;
  }
  impulse_mutex_destroy(queue->mutex);
  impulse_cond_destroy(queue->condition);
  impulse_free(queue);
}


void *impulse_message_queue_receive(impulse_message_queue queue)
{
  void *data;
  struct _p_impulse_thread_message *next;

  impulse_mutex_lock(queue->mutex);

  while (queue->first == NULL)
    impulse_cond_wait(queue->condition, queue->mutex);

  /* We have a message - retrieve the data */
  data = queue->first->data;
  next = queue->first->next;

  /* Free the queue & move along a message */
  if (next == NULL)
    queue->last = NULL;
  
  impulse_free(queue->first);
  queue->first = next;

  impulse_mutex_unlock(queue->mutex);

  return data;
}


impulse_bool impulse_message_queue_full(impulse_message_queue queue)
{
  impulse_bool result;

  result = 0;

  if (queue->first != NULL)
    result = 1;

  return result;
}

void *impulse_message_queue_getnext(impulse_message_queue queue)
{
  void *data;
  struct _p_impulse_thread_message *next;

  if (queue->first == NULL)
    return NULL;

  /* We have a message - retrieve the data */
  data = queue->first->data;
  next = queue->first->next;

  /* Free the queue & move along a message */
  if (next == NULL)
    queue->last = NULL;
  
  impulse_free(queue->first);
  queue->first = next;

  return data;
}


void impulse_message_queue_send(impulse_message_queue queue,
				void *data)
{
  struct _p_impulse_thread_message *next;

  impulse_mutex_lock(queue->mutex);

  /* Add the new item to the queue */
  next = impulse_alloc(sizeof(struct _p_impulse_thread_message));

  next->data = data;
  next->next = NULL;

  if (queue->last == NULL)
  {
    queue->first = next;
    queue->last = next;
  }
  else
  {
    queue->last->next = next;
    queue->last = next;
  }

  impulse_mutex_unlock(queue->mutex);

  /* Signal the items availability */
  impulse_cond_broadcast(queue->condition);
}


void impulse_message_queue_lock(impulse_message_queue queue)
{
  impulse_mutex_lock(queue->mutex);
}

void impulse_message_queue_unlock(impulse_message_queue queue)
{
  impulse_mutex_unlock(queue->mutex);
}

#endif


